'SBMIDCHK
'Stand-alone SBMIDI driver detection utility
'by Jesse Dorland (jessedorland@hotmail.com)
'Released to the public domain, June 1998.
'
'NOTE: SBMIDCHK.BAS will not run in source form with any flavor of
'BASIC.  It must be compiled to work properly.  A compiled version
'(SBMIDCHK.EXE) is provided.
'
'SBMIDCHK is a stand-alone utility for detecting the SBMIDI sound
'driver.  The source code is of no use, but when compiled, SBMIDCHK
'provides a way to detect the driver (and load it, if necessary)
'before your program starts, using batch files.  This is handy because
'the SBMIDI and SBSIM drivers cannot be loaded properly from any running
'BASIC program.
'
'SBMIDCHK returns an errorlevel.  If it returns 0, it didn't locate SBMIDI.
'If it did locate the driver, it returns a non-zero value, equal to the
'interrupt number that the driver is using.
'
'The SBSIMCHK utility works just like SBMIDCHK, except that it detects the
'SBSIM driver.

DECLARE FUNCTION SBMIDILoaded% ()
DECLARE SUB GetIntVector (IntNum%, Segment%, Offset%)
DECLARE SUB EndWithErrorLevel ALIAS "_Exit" (BYVAL ErrorLevel%)

'Check if the drivers are loaded
ErrorLevel% = SBMIDILoaded%

'Exit with the proper errorlevel
EndWithErrorLevel ErrorLevel%

DEFINT A-Z
SUB GetIntVector (IntNum%, Segment%, Offset%) STATIC
'If the code hasn't been loaded already, do it now.
IF GetIntVCodeLoaded% = 0 THEN
    asm$ = asm$ + CHR$(&H55)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HE5)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&HA)
    asm$ = asm$ + CHR$(&H8A) + CHR$(&H7)
    asm$ = asm$ + CHR$(&HB4) + CHR$(&H35)
    asm$ = asm$ + CHR$(&HCD) + CHR$(&H21)
    asm$ = asm$ + CHR$(&H8C) + CHR$(&HC1)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HDA)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H8)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HF)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H6)
    asm$ = asm$ + CHR$(&H89) + CHR$(&H17)
    asm$ = asm$ + CHR$(&H5D)
    asm$ = asm$ + CHR$(&HCB)
    asm$ = asm$ + CHR$(&H34) + CHR$(&H0)
    asm$ = asm$ + CHR$(&H60)
    asm$ = asm$ + CHR$(&H23) + CHR$(&H0)
    GetIntVCodeLoaded% = 1
END IF
'Execute the code
DEF SEG = VARSEG(asm$)
CALL ABSOLUTE(IntNum%, Segment%, Offset%, SADD(asm$))
END SUB

FUNCTION SBMIDILoaded% STATIC
'Open the data file.
FF% = FREEFILE
OPEN "DRIVERS.DAT" FOR BINARY AS #FF%
'If the file is empty, return an error
FileSize& = LOF(FF%)
IF FileSize& = 0 THEN
    CLOSE FF%
    KILL "DRIVERS.DAT"
    MIDI.ERROR = 1
    EXIT FUNCTION
'If the file is not exactly 1 kilobyte in size, return an error
ELSEIF FileSize& <> 1024 THEN
    CLOSE FF%
    MIDI.ERROR = 9
    EXIT FUNCTION
END IF
'Load the driver data.
DIM DRIVERDATA$(1 TO 5)
FOR I% = 1 TO 4
    DRIVERDATA$(I%) = INPUT$(256, #FF%)
NEXT I%
CLOSE #FF%

'Check the interrupt handlers for int 80h-FFh to see if they contain
'program code from SBMIDI.
FOR I% = &H80 TO &HFF
    'Get the address of the interrupt handler.
    GetIntVector I%, Segment%, Offset%
    'If the address is null, then the interrupt is not in use, and can be
    'skipped.
    IF Segment% = 0 AND Offset% = 0 THEN GOTO Skip:
    'The following code checks for the driver by looking for the text
    '"SBMIDI" at a certain location in the driver code.  It sounds
    'simplistic, but it's very accurate.  If it doesn't work, a different
    'method is used.
    IF SBMIDI% = 0 THEN
        NewSegment% = CVI(MKI$(CVL(MKI$(Segment%) + CHR$(0) + CHR$(0)) - &H11&))
        DEF SEG = NewSegment%
        TEMP$ = ""
        FOR J% = 1 TO 6
            TEMP$ = TEMP$ + CHR$(PEEK(271 + J%))
        NEXT
        IF TEMP$ = "SBMIDI" THEN SBMIDILoaded% = I%: EXIT FUNCTION
    END IF
    'Point to the segment of the interrupt handler.
    DEF SEG = Segment%
    'Read 256 bytes of the driver code.
    DRIVERDATA$(5) = ""
    FOR J% = 0 TO 255
        Byte% = PEEK(Offset% + J%)
        DRIVERDATA$(5) = DRIVERDATA$(5) + CHR$(Byte%)
    NEXT J%
    'Check to see if the code matches any of the previously saved data.
    FOR J% = 2 TO 4
        MATCH% = 1
        FOR K% = 0 TO 255
            IF MID$(DRIVERDATA$(J%), K% + 1, 1) <> MID$(DRIVERDATA$(5), K% + 1, 1) THEN
                SELECT CASE K%
                    CASE IS = 14, 15, 113, 114, 235, 236
                    CASE ELSE
                        MATCH% = 0
                        EXIT FOR
                END SELECT
            END IF
        NEXT K%
        IF MATCH% THEN SBMIDILoaded% = I%: EXIT FUNCTION
    NEXT J%
Skip:
NEXT I%
END FUNCTION

