'SBSIMCHK v1.0
'Stand-alone SBSIM driver detection utility
'by Jesse Dorland (jessedorland@hotmail.com)
'Released to the public domain, June 1998.
'
'NOTE: SBSIMCHK.BAS will not run in source form with any flavor of
'BASIC.  It must be compiled to work properly.  A compiled version
'(SBSIMCHK.EXE) is provided.
'
'SBSIMCHK is a stand-alone utility for detecting the SBSIM sound
'driver.  The source code is of no use, but when compiled, SBSIMCHK
'provides a way to detect the driver (and load it, if necessary)
'before your program starts, using batch files.  This is handy because,
'while you can check to see if the drivers are loaded while your program
'is running, you can't actually load the drivers.
'
'SBSIMCHK returns an errorlevel; if it returns a non-zero value, it found
'driver.  If it returns 0, it didn't locate SBSIM.
'
'The SBMIDCHK utility works just like SBSIMCHK, except that it detects the
'SBMIDI driver.

DECLARE FUNCTION SBSIMLoaded% ()
DECLARE SUB GetIntVector (IntNum%, Segment%, Offset%)
DECLARE SUB EndWithErrorLevel ALIAS "_Exit" (BYVAL ErrorLevel%)

'Check if the drivers are loaded
ErrorLevel% = SBSIMLoaded%

'Exit with the proper errorlevel
EndWithErrorLevel ErrorLevel%

DEFINT A-Z
SUB GetIntVector (IntNum%, Segment%, Offset%) STATIC
'If the code hasn't been loaded already, do it now.
IF GetIntVCodeLoaded% = 0 THEN
    asm$ = asm$ + CHR$(&H55)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HE5)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&HA)
    asm$ = asm$ + CHR$(&H8A) + CHR$(&H7)
    asm$ = asm$ + CHR$(&HB4) + CHR$(&H35)
    asm$ = asm$ + CHR$(&HCD) + CHR$(&H21)
    asm$ = asm$ + CHR$(&H8C) + CHR$(&HC1)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HDA)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H8)
    asm$ = asm$ + CHR$(&H89) + CHR$(&HF)
    asm$ = asm$ + CHR$(&H8B) + CHR$(&H5E) + CHR$(&H6)
    asm$ = asm$ + CHR$(&H89) + CHR$(&H17)
    asm$ = asm$ + CHR$(&H5D)
    asm$ = asm$ + CHR$(&HCB)
    asm$ = asm$ + CHR$(&H34) + CHR$(&H0)
    asm$ = asm$ + CHR$(&H60)
    asm$ = asm$ + CHR$(&H23) + CHR$(&H0)
    GetIntVCodeLoaded% = 1
END IF
'Execute the code
DEF SEG = VARSEG(asm$)
CALL ABSOLUTE(IntNum%, Segment%, Offset%, SADD(asm$))
END SUB

FUNCTION SBSIMLoaded% STATIC
'Open the data file.
FF% = FREEFILE
OPEN "DRIVERS.DAT" FOR BINARY AS #FF%
'If the file is empty, return an error
FileSize& = LOF(FF%)
IF FileSize& = 0 THEN
    CLOSE FF%
    KILL "DRIVERS.DAT"
    MIDI.ERROR = 1
    EXIT FUNCTION
'If the file is not exactly 1 kilobyte in size, return an error
ELSEIF FileSize& <> 1024 THEN
    CLOSE FF%
    MIDI.ERROR = 9
    EXIT FUNCTION
END IF
'Load the driver data.
DIM DRIVERDATA$(1 TO 5)
DRIVERDATA$(1) = INPUT$(256, #FF%)
CLOSE #FF%

'Check the interrupt handlers for int 80h-FFh to see if they contain
'program code from SBMIDI.
FOR I% = &H80 TO &HFF
    'Get the address of the interrupt handler.
    GetIntVector I%, Segment%, Offset%
    'If the address is null, then the interrupt is not in use, and can be
    'skipped.
    IF Segment% = 0 AND Offset% = 0 THEN GOTO Skip:
    IF SBSIM% = 0 THEN
        NewSegment% = CVI(MKI$(CVL(MKI$(Segment%) + CHR$(0) + CHR$(0)) - &H1&))
        DEF SEG = NewSegment%
        TEMP$ = ""
        FOR J% = 1 TO 5
            TEMP$ = TEMP$ + CHR$(PEEK(274 + J%))
        NEXT
        IF TEMP$ = "SBSIM" THEN SBSIMLoaded% = I%: EXIT FUNCTION
    END IF
    'Point to the segment of the interrupt handler.
    DEF SEG = Segment%
    'Read 256 bytes of the driver code.
    DRIVERDATA$(5) = ""
    FOR J% = 0 TO 255
        Byte% = PEEK(Offset% + J%)
        DRIVERDATA$(5) = DRIVERDATA$(5) + CHR$(Byte%)
    NEXT J%
    'Check to see if the code matches any of the previously saved data.
    MATCH% = 1
    FOR K% = 0 TO 255
        IF MID$(DRIVERDATA$(1), K% + 1, 1) <> MID$(DRIVERDATA$(5), K% + 1, 1) THEN
            SELECT CASE K%
                CASE IS = 14, 15, 113, 114, 235, 236
                CASE ELSE
                    MATCH% = 0
                    EXIT FOR
            END SELECT
        END IF
    NEXT K%
    IF MATCH% THEN SBSIMLoaded% = I%: EXIT FUNCTION
Skip:
NEXT I%
END FUNCTION

